/*
 * xinit.c: Initialize XView, create XView-objects 
 *
 * Copyright (c) 1994-96 Martin Buck  <martin-2.buck@student.uni-ulm.de>
 * Read COPYING for more information
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <xview/xview.h>
#include <xview/frame.h>
#include <xview/cms.h>
#include <xview/panel.h>
#include <xview/canvas.h>
#include <xview/notify.h>
#include <xview/svrimage.h>
#include <xview/icon.h>
#ifdef NEED_FULLSCREEN_WORKAROUND
#include <xview/fullscreen.h>
#endif
#include "safe_malloc.h"
#include "tilde.h"
#include "misc.h"
#include "cct.h"
#include "vtxdecode.h"
#include "vtxqueue.h"
#include "config.h"
#include "dialog.h"
#include "xevents.h"
#include "xinit.h"

#include "vtx.xbm"
#include "empty.xbm"
#include "eyes.xbm"
#include "sheet.xbm"
#include "checkmark_sheet.xbm"
#include "checkmark_sheets.xbm"
#include "clock.xbm"
#include "eyes_sheet.xbm"
#include "sheets.xbm"

#define VTXRC_FNAME "vtxrc"
#define MAXFONTS 8

char *intv_fname;
Display *dpy;
Window rootid, frameid;
int screen_num;
unsigned long vtxcolors[8];
GC vtxgc;
Pixmap dith_bm[8];
vtxfont_t vtxfonts[MAXFONTS];
int vtxmaxfont;
Frame frame;
Canvas canvas, canvas_bg;
Panel right_panel, main_panel, hist_panel;
Panel_item pgnum_item, subpgnum_item, cont_button, checkbox, intv_reveal_toggle, pgmsg_item;
Server_image hist_images[STAT_COUNT];
Menu scr_menu, font_menu;
vtx_info_t vtx_info;



static void
save_user_config(void) {
  char *rcfile, *tmprcfile;
  
  config_set_font(vtxwin.font);
  rcfile = sstrdup("~/." VTXRC_FNAME);
  if (tilde_expand_realloc(&rcfile) >= 0) {
    tmprcfile = sstrdup(rcfile);
    tmprcfile = sstrapp(tmprcfile, ".tmp");
    remove(tmprcfile);
    if (config_save(tmprcfile) || remove(rcfile) < 0 || rename(tmprcfile, rcfile) < 0) {
      confirm_notice_v("Error: Can't save options to ", rcfile, ":\n", strerror(errno), NULL);
    }
    remove(tmprcfile);
    free(tmprcfile);
  } else {
    confirm_notice("Error: Can't save options: No home directory found.");
  }
  free(rcfile);
}


int
main(int argc, char *argv[]) {
  char **fontlist, *rcfile, tmpstr[256];
  int optchr, font, fontnum, item;
  Menu file_menu, gadg_menu, opt_menu;
  Cms cms;
  Server_image icon_image;
  Icon icon;
  const Xv_singlecolor rgb_val[8] = {
    { 255, 255, 255 },	/* white */
    {   0, 255, 255 },	/* cyan */
    { 255,   0, 255 },	/* magenta */
    {   0,   0, 255 },	/* blue */
    { 255, 255,   0 },	/* yellow */
    {   0, 255,   0 },	/* green */
    { 255,   0,   0 },	/* red */
    {   0,   0,   0 },	/* black */
  };
  const char dith_data[8][8] = {
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },	/* black */
    { 0x00, 0x33, 0x00, 0x33, 0x00, 0x33, 0x00, 0x33 },	/* red */
    { 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa },	/* green */
    { 0x66, 0xff, 0x66, 0xff, 0x66, 0xff, 0x66, 0xff },	/* yellow */
    { 0x55, 0x00, 0x55, 0x00, 0x55, 0x00, 0x55, 0x00 },	/* blue */
    { 0x33, 0xcc, 0x33, 0xcc, 0x33, 0xcc, 0x33, 0xcc }, /* magenta */
    { 0xff, 0x33, 0xff, 0x33, 0xff, 0x33, 0xff, 0x33 },	/* cyan */
    { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }	/* white */
  };

  if ((smalloc_progname = strrchr(argv[0], '/'))) {
    smalloc_progname++;
  } else {
    smalloc_progname = argv[0];
  }

  xv_init(XV_INIT_ARGC_PTR_ARGV, &argc, argv, NULL);
  
  frame = (Frame)xv_create(XVNULL, FRAME,
      WIN_EVENT_PROC, frame_proc,
      WIN_CONSUME_EVENTS,
        WIN_STRUCTURE_NOTIFY, WIN_PROPERTY_NOTIFY, NULL,
      FRAME_LABEL, VTXWINNAME " V" VTXVERSION,
      FRAME_SHOW_FOOTER, TRUE,
      NULL);

#ifdef NEED_FULLSCREEN_WORKAROUND
  xv_destroy((Fullscreen)xv_create(frame, FULLSCREEN, NULL));
#endif

  dpy = (Display*)xv_get(frame, XV_DISPLAY);
  rootid = (Window)xv_get((Xv_object)xv_get(frame, XV_ROOT), XV_XID);
  screen_num = (int)xv_get((Xv_screen)xv_get(frame, XV_SCREEN), SCREEN_NUMBER);
  frameid = (Window)xv_get(frame, XV_XID);

  cms = (Cms)xv_create(XVNULL, CMS,
      CMS_SIZE, 8,
      CMS_COLORS, rgb_val,
      NULL);
  if ((Colormap)xv_get(cms, XV_XID) != DefaultColormap(dpy, screen_num)) {
    cms = (Cms)xv_create(XVNULL, CMS,
        CMS_SIZE, 8,
        CMS_NAMED_COLORS, "white", "black", "black", "black", "black", "black", "black", "black", NULL,
        NULL);
    do_dither = TRUE;
  } else if ((int)xv_get(frame, WIN_DEPTH) < 3) {
    do_dither = TRUE;
  }

  for (item = 0; item <= 7; item++)
    vtxcolors[7 - item] = (unsigned long)xv_get(cms, CMS_PIXEL, item);

  vtxgc = XCreateGC(dpy, rootid, 0, NULL);
  for (item = 0; item <= 7; item++) {
    dith_bm[item] = XCreateBitmapFromData(dpy, rootid, dith_data[item], 8, 8);
  }
			/* Find all fonts that match the following pattern & get their font-info */
  fontlist = XListFonts(dpy, "-misc-videotext-medium-r-normal--*-*-75-75-c-*-misc-fontspecific",
      MAXFONTS + 1, &vtxmaxfont);
  fontnum = 0;
  for (font = 0; fontlist && font < vtxmaxfont && fontnum < MAXFONTS; font++) {
    if (strncmp(fontlist[font], "-misc-videotext-medium-r-normal--0-", 35)) {
      if ((vtxfonts[fontnum].normal = XLoadQueryFont(dpy, fontlist[font]))) {
        vtxfonts[fontnum].height = vtxfonts[fontnum].normal->max_bounds.ascent;

        sprintf(tmpstr, "-misc-videotext-medium-r-half--%d-*-75-75-c-*-misc-fontspecific",
            vtxfonts[fontnum].height * 2);
        if ((vtxfonts[fontnum].doubleht = XLoadQueryFont(dpy, tmpstr))) {
          vtxfonts[fontnum].width = vtxfonts[fontnum].normal->max_bounds.width;
          sprintf(vtxfonts[fontnum].fontname, "%dx%d", (unsigned char)vtxfonts[fontnum].width,
              (unsigned char)vtxfonts[fontnum].height);
          fontnum++;
        } else {
          fprintf(stderr, "Warning: can't load double-height-font (%s)\n", tmpstr);
        }
      }
    }
  }
  if (!fontnum) {
    fprintf(stderr, "Error: No VideoteXt-fonts found "
        "(-misc-videotext-medium-r-normal--*-*-75-75-c-*-misc-fontspecific)\n");
    exit(1);
  }
  vtxmaxfont = fontnum - 1;
  XFreeFontNames(fontlist);

  /* We can't load the config-files earlier, since the font-names must be already known */
  config_init();
  if (config_load(VTX_LIBDIR "/" VTXRC_FNAME, FALSE) && errno != ENOENT) {
    fprintf(stderr, "%s: " VTX_LIBDIR "/" VTXRC_FNAME ": %s\n", smalloc_progname, strerror(errno));
    exit(1);
  }

  rcfile = sstrdup("~/." VTXRC_FNAME);
  if (tilde_expand_realloc(&rcfile) >= 0) {
    if (config_load(rcfile, TRUE) && errno != ENOENT) {
      fprintf(stderr, "%s: %s: %s\n", smalloc_progname, rcfile, strerror(errno));
      exit(1);
    }
  } else {
    fprintf(stderr, "%s: Warning: Can't load user-defaults: No home directory found.\n",
        smalloc_progname);
  }
  free(rcfile);

  if (getenv("VTX_DEVICE")) {
    cct_device = getenv("VTX_DEVICE");
  }
  if (argc > 1) {
    while ((optchr = getopt(argc, argv, "f:D:P:?")) != -1) {
      switch (optchr) {
        case 'f':
          intv_fname = optarg;
        break;
        case 'D':
          cct_device = optarg;
        break;
        case 'P':
          vtx_fontpath = optarg;
        break;
        case '?':
          printf("Usage: %s [standard XView-options] [options]\n"
                 "Options:\n"
                 "  -f <file>    Load file & enter display-only mode\n"
                 "  -D <device>  Set device to use (default: /dev/vtx)\n"
                 "  -P <path>    Set font-path for videotext-fonts (not X-fonts!)\n"
                 "  -\\?          Show this help\n"
                 "  -help        Show XView-options\n", smalloc_progname);
          exit(1);
        break;
      }
    }
  }

  icon_image = (Server_image)xv_create(XVNULL, SERVER_IMAGE,
      XV_WIDTH, vtx_width,
      XV_HEIGHT, vtx_height,
      SERVER_IMAGE_X_BITS, vtx_bits,
      NULL);
  icon = (Icon)xv_create(frame, ICON,
      ICON_IMAGE, icon_image,
      NULL);
  hist_images[STAT_NEW] = (Server_image)xv_create(XVNULL, SERVER_IMAGE,
      XV_WIDTH, empty_width,
      XV_HEIGHT, empty_height,
      SERVER_IMAGE_X_BITS, empty_bits,
      NULL);
  hist_images[STAT_SEARCH] = (Server_image)xv_create(XVNULL, SERVER_IMAGE,
      XV_WIDTH, eyes_width,
      XV_HEIGHT, eyes_height,
      SERVER_IMAGE_X_BITS, eyes_bits,
      NULL);
  hist_images[STAT_FOUND] = (Server_image)xv_create(XVNULL, SERVER_IMAGE,
      XV_WIDTH, sheet_width,
      XV_HEIGHT, sheet_height,
      SERVER_IMAGE_X_BITS, sheet_bits,
      NULL);
  hist_images[STAT_READ] = (Server_image)xv_create(XVNULL, SERVER_IMAGE,
      XV_WIDTH, checkmark_sheet_width,
      XV_HEIGHT, checkmark_sheet_height,
      SERVER_IMAGE_X_BITS, checkmark_sheet_bits,
      NULL);
  hist_images[STAT_READ_MP] = (Server_image)xv_create(XVNULL, SERVER_IMAGE,
      XV_WIDTH, checkmark_sheets_width,
      XV_HEIGHT, checkmark_sheets_height,
      SERVER_IMAGE_X_BITS, checkmark_sheets_bits,
      NULL);
  hist_images[STAT_TIMEOUT] = (Server_image)xv_create(XVNULL, SERVER_IMAGE,
      XV_WIDTH, clock_width,
      XV_HEIGHT, clock_height,
      SERVER_IMAGE_X_BITS, clock_bits,
      NULL);
  hist_images[STAT_FOUND_MP] = (Server_image)xv_create(XVNULL, SERVER_IMAGE,
      XV_WIDTH, sheets_width,
      XV_HEIGHT, sheets_height,
      SERVER_IMAGE_X_BITS, sheets_bits,
      NULL);
  hist_images[STAT_SEARCH_MP] = (Server_image)xv_create(XVNULL, SERVER_IMAGE,
      XV_WIDTH, eyes_sheet_width,
      XV_HEIGHT, eyes_sheet_height,
      SERVER_IMAGE_X_BITS, eyes_sheet_bits,
      NULL);

  xv_set(frame,
      FRAME_MIN_SIZE, 41 * vtxfonts[0].width + (intv_fname ? 0 : RPANEL_WIDTH) +
          4 * WIN_DEFAULT_BORDER_WIDTH, 26 * vtxfonts[0].height + CANVAS_OFFSET,
      FRAME_MAX_SIZE, 41 * vtxfonts[vtxmaxfont].width + (intv_fname ? 0 : RPANEL_WIDTH) +
          4 * WIN_DEFAULT_BORDER_WIDTH, 26 * vtxfonts[vtxmaxfont].height + CANVAS_OFFSET,
      FRAME_ICON, icon,
      NULL);

  main_panel = (Panel)xv_create(frame, PANEL,
      XV_X, 0,
      XV_Y, 0,
      XV_HELP_DATA, "videotext:main_panel",
      NULL);

  file_menu = (Menu)xv_create(XVNULL, MENU,
      MENU_GEN_PIN_WINDOW, frame, "File",
      MENU_ACTION_ITEM, "Load VTX...", load_proc,
      MENU_ACTION_ITEM, "Save VTX...", save_proc,
      MENU_ACTION_ITEM, "Export...", export_proc,
      MENU_ITEM, MENU_STRING, "", MENU_INACTIVE, TRUE, NULL,
      NULL);
  if (!intv_fname) {
    xv_set(file_menu,
        MENU_ACTION_ITEM, "Open device", open_device,
        MENU_ACTION_ITEM, "Close device", close_device,
        MENU_ACTION_ITEM, "Reset", reset_proc,
        MENU_ITEM, MENU_STRING, "", MENU_INACTIVE, TRUE, NULL,
        MENU_ACTION_ITEM, "Statistics...", stat_proc,
        NULL);
  }
  xv_set(file_menu,
    MENU_ACTION_ITEM, "About...", about_proc,
    MENU_ITEM, MENU_STRING, "", MENU_INACTIVE, TRUE, NULL,
    MENU_ACTION_ITEM, "Exit", xquit,
    NULL);
  xv_create(main_panel, PANEL_BUTTON,
      PANEL_LABEL_STRING, "File",
      PANEL_ITEM_MENU, file_menu,
      XV_X, 10,
      XV_Y, 10,
      NULL);

  gadg_menu = (Menu)xv_create(XVNULL, MENU,
      MENU_GEN_PIN_WINDOW, frame, "Gadgets",
      MENU_ACTION_ITEM, "Page history...", history_proc,
      MENU_ACTION_ITEM, "TOP-Text...", toptext_proc,
      MENU_ACTION_ITEM, "Hotlist...", hotlist_proc,
      MENU_ACTION_ITEM, "Select station...", sel_station_proc,
      NULL);
  if (!intv_fname) {
    xv_create(main_panel, PANEL_BUTTON,
        PANEL_LABEL_STRING, "Gadgets",
        PANEL_ITEM_MENU, gadg_menu,
        XV_X, 80,
        XV_Y, 10,
        NULL);
  }

  scr_menu = (Menu)xv_create(XVNULL, MENU_CHOICE_MENU, NULL);
  for (item = 0; item < update_table_count; item++) {
    xv_set(scr_menu, MENU_APPEND_ITEM,
        (Menu_item)xv_create(XVNULL, MENUITEM,
            MENU_STRING, update_table[item].menu_string,
            MENU_NOTIFY_PROC, update_menu_proc,
            MENU_CLIENT_DATA, item,
            MENU_SELECTED, (item == 1),
            MENU_RELEASE,
            NULL),
        NULL);
  }
  font_menu = (Menu)xv_create(XVNULL, MENU_CHOICE_MENU, NULL);
  for (item = 0; item <= vtxmaxfont; item++) {
    xv_set(font_menu, MENU_APPEND_ITEM,
        (Menu_item)xv_create(XVNULL, MENUITEM,
            MENU_STRING, vtxfonts[item].fontname,
            MENU_NOTIFY_PROC, font_menu_proc,
            MENU_CLIENT_DATA, item,
            MENU_RELEASE,
            NULL),
        NULL);
  }
  opt_menu = (Menu)xv_create(XVNULL, MENU,
      MENU_GEN_PIN_WINDOW, frame, "Options",
      MENU_PULLRIGHT_ITEM, "Font", font_menu,
      MENU_PULLRIGHT_ITEM, "Screen updates", scr_menu,
      MENU_ACTION_ITEM, "Window layout...", layout_proc,
      MENU_ACTION_ITEM, "TV options...", tvopt_proc,
      MENU_ACTION_ITEM, "Misc. options...", miscopt_proc,
      MENU_ITEM, MENU_STRING, "", MENU_INACTIVE, TRUE, NULL,
      MENU_ACTION_ITEM, "Save global options", save_user_config,
      MENU_ITEM, MENU_STRING, "", MENU_INACTIVE, TRUE, NULL,
      MENU_ACTION_ITEM, "Station options...", station_opt_proc,
      MENU_ACTION_ITEM, "Spool options...", spool_opt_proc,
      NULL);
  xv_create(main_panel, PANEL_BUTTON,
      PANEL_LABEL_STRING, intv_fname ? "Fonts" : "Options",
      PANEL_ITEM_MENU, intv_fname ? font_menu : opt_menu,
      XV_X, intv_fname ? 80 : 180,
      XV_Y, 10,
      NULL);
  intv_reveal_toggle = (Panel_item)xv_create(main_panel, PANEL_TOGGLE,
      XV_Y, 8,
      PANEL_CHOICE_STRINGS, "Reveal", NULL,
      PANEL_NOTIFY_PROC, intv_reveal_proc,
      XV_SHOW, !!intv_fname,
      NULL);

  right_panel = (Panel)xv_create(frame, PANEL,
      XV_X, 41 * vtxfonts[0].width + 4 * WIN_DEFAULT_BORDER_WIDTH,
      XV_Y, CANVAS_OFFSET,
      PANEL_ACCEPT_KEYSTROKE, TRUE,
      XV_HELP_DATA, "videotext:main_panel",
      NULL);
  xv_create(right_panel, PANEL_MESSAGE,
      XV_X, 10,
      XV_Y, 0,
      PANEL_LABEL_STRING, "Current:",
      PANEL_LABEL_BOLD, TRUE,
      NULL);
  pgmsg_item = (Panel_item)xv_create(right_panel, PANEL_MESSAGE,
      XV_X, 10,
      XV_Y, 15,
      PANEL_LABEL_STRING, "?/?/?",
      NULL);
  pgnum_item = (Panel_item)xv_create(right_panel, PANEL_TEXT,
      XV_X, 10,
      XV_Y, 35,
      PANEL_LABEL_STRING, "Page:",
      PANEL_VALUE_DISPLAY_WIDTH, 35,
      PANEL_VALUE_STORED_LENGTH, 3,
      PANEL_NOTIFY_STRING, "\n\r",
      PANEL_NOTIFY_PROC, pgnum_proc,
      XV_SHOW, !intv_fname,
      NULL);
  xv_create(right_panel, PANEL_BUTTON,
      XV_X, 10,
      XV_Y, 55,
      PANEL_LABEL_STRING, " +",
      PANEL_LABEL_WIDTH, 20,
      PANEL_NOTIFY_PROC, pginc_proc,
      XV_SHOW, !intv_fname,
      NULL);
  xv_create(right_panel, PANEL_BUTTON,
      XV_X, 50,
      XV_Y, 55,
      PANEL_LABEL_STRING, " -",
      PANEL_LABEL_WIDTH, 20,
      PANEL_NOTIFY_PROC, pgdec_proc,
      XV_SHOW, !intv_fname,
      NULL);
  subpgnum_item = (Panel_item)xv_create(right_panel, PANEL_TEXT,
      XV_X, 10,
      XV_Y, 80,
      PANEL_LABEL_STRING, "Subpg.:",
      PANEL_VALUE_DISPLAY_WIDTH, 23,
      PANEL_VALUE_STORED_LENGTH, 2,
      PANEL_NOTIFY_STRING, "\n\r",
      PANEL_NOTIFY_PROC, pgnum_proc,
      XV_SHOW, !intv_fname,
      NULL);
  xv_create(right_panel, PANEL_BUTTON,
      XV_X, 10,
      XV_Y, 100,
      PANEL_LABEL_STRING, " +",
      PANEL_LABEL_WIDTH, 20,
      PANEL_NOTIFY_PROC, subpginc_proc,
      XV_SHOW, !intv_fname,
      NULL);
  xv_create(right_panel, PANEL_BUTTON,
      XV_X, 50,
      XV_Y, 100,
      PANEL_LABEL_STRING, " -",
      PANEL_LABEL_WIDTH, 20,
      PANEL_NOTIFY_PROC, subpgdec_proc,
      XV_SHOW, !intv_fname,
      NULL);
  xv_create(right_panel, PANEL_BUTTON,
      XV_X, 10,
      XV_Y, 130,
      PANEL_LABEL_STRING, "    Roll",
      PANEL_LABEL_WIDTH, 60,
      PANEL_NOTIFY_PROC, subpgroll_proc,
      XV_SHOW, !intv_fname,
      NULL);
  xv_create(right_panel, PANEL_BUTTON,
      XV_X, 10,
      XV_Y, 165,
      PANEL_LABEL_STRING, "  Search",
      PANEL_LABEL_WIDTH, 60,
      PANEL_NOTIFY_PROC, pgnum_proc,
      XV_SHOW, !intv_fname,
      NULL);
  xv_create(right_panel, PANEL_BUTTON,
      XV_X, 10,
      XV_Y, 195,
      PANEL_LABEL_STRING, "Search bg",
      PANEL_LABEL_WIDTH, 60,
      PANEL_NOTIFY_PROC, pgnum_bg_proc,
      XV_SHOW, !intv_fname,
      NULL);
  cont_button = (Panel_item)xv_create(right_panel, PANEL_TOGGLE,
      XV_X, 10,
      XV_Y, 225,
      PANEL_CHOICE_STRINGS, " Continue", NULL,
      XV_SHOW, !intv_fname,
      NULL);
  checkbox = (Panel_item)xv_create(right_panel, PANEL_CHECK_BOX,
      XV_X, 10,
      XV_Y, 255,
      PANEL_LAYOUT, PANEL_VERTICAL,
      PANEL_CHOICE_STRINGS, "Stop", "Reveal", NULL,
      PANEL_NOTIFY_PROC, checkbox_proc,
      XV_SHOW, !intv_fname,
      NULL);

  canvas_bg = (Canvas)xv_create(frame, CANVAS,
      XV_X, 0,
      XV_Y, CANVAS_OFFSET,
      CANVAS_AUTO_SHRINK, FALSE,
      CANVAS_AUTO_EXPAND, FALSE,
      WIN_CMS, cms,
      WIN_BACKGROUND_COLOR, CMSCBLACK,
      WIN_FOREGROUND_COLOR, CMSCWHITE,
      NULL);
  canvas = (Canvas)xv_create(canvas_bg, CANVAS,
      XV_X, 1,
      XV_Y, vtxfonts[0].height - 2 * WIN_DEFAULT_BORDER_WIDTH,
      CANVAS_AUTO_SHRINK, FALSE,
      CANVAS_AUTO_EXPAND, FALSE,
      CANVAS_REPAINT_PROC, canvas_repaint,
      CANVAS_X_PAINT_WINDOW, FALSE,
      CANVAS_FIXED_IMAGE, FALSE,
      WIN_CMS, cms,
      WIN_BACKGROUND_COLOR, CMSCBLACK,
      WIN_FOREGROUND_COLOR, CMSCBLACK,
      WIN_EVENT_PROC, canvas_proc,
      WIN_CONSUME_EVENTS,
        WIN_MOUSE_BUTTONS, KBD_USE, NULL,
      NULL);
  xv_set(canvas_paint_window(canvas),
      WIN_EVENT_PROC, canvas_proc,
      WIN_CONSUME_EVENTS,
        WIN_MOUSE_BUTTONS, NULL,
      NULL);

  popups[POPUP_HISTORY].frame = (Frame)xv_create(frame, FRAME,
      XV_LABEL, "Page history",
      NULL);
  hist_panel = (Panel)xv_create(popups[POPUP_HISTORY].frame, PANEL,
      PANEL_ACCEPT_KEYSTROKE, TRUE,
      XV_HELP_DATA, "videotext:history",
      NULL);
  popups[POPUP_HISTORY].list = (Panel_item)xv_create(hist_panel, PANEL_LIST,
      PANEL_READ_ONLY, TRUE,
      NULL);

  notify_interpose_event_func(right_panel, check_hotkeys, NOTIFY_SAFE);

  /* required for popen() in print_ps */
  notify_set_signal_func(frame, signal_handler, SIGPIPE, NOTIFY_SYNC);
  notify_set_signal_func(frame, signal_handler, SIGUSR1, NOTIFY_SYNC);
  x_main_loop();
  exit(0);
}
